//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/HardParticle/HollowSphere.cpp
//! @brief     Implements class HollowSphere.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/HardParticle/HollowSphere.h"
#include "Base/Math/Constants.h"
#include "Base/Util/Assert.h"
#include "Sample/Shapes/TruncatedEllipsoidNet.h"
#include <limits>

HollowSphere::HollowSphere(const std::vector<double> P)
    : IFormFactor(P)
    , m_mean(m_P[0])
    , m_full_width(m_P[1])
{
    validateOrThrow();
}

HollowSphere::HollowSphere(double mean, double full_width)
    : HollowSphere(std::vector<double>{mean, full_width})
{
}

complex_t HollowSphere::formfactor_at_bottom(C3 q) const
{
    ASSERT(m_validated);
    double R = m_mean;
    double W = m_full_width;
    double q2 = std::norm(q.x()) + std::norm(q.y()) + std::norm(q.z());
    double q_r = std::sqrt(q2);
    if (q_r * R < std::numeric_limits<double>::epsilon())
        return (4.0 * M_PI * R * R * R + M_PI * R * W * W) / 3.0;
    double qR = q_r * R;
    double qW = q_r * W;
    double numerator =
        4 * M_PI
        * (4 * std::sin(qR) * std::sin(qW / 2.0) - qW * std::cos(qW / 2.0) * std::sin(qR)
           - 2.0 * qR * std::cos(qR) * std::sin(qW / 2.0));
    return numerator / (q2 * q2 * W);
}

std::string HollowSphere::validate() const
{
    std::vector<std::string> errs;
    requestGt0(errs, m_mean, "mean radius");
    requestGe0(errs, m_full_width, "full width");
    if (2 * m_mean < m_full_width)
        errs.push_back("parameters violate condition full_width<=2R");
    if (!errs.empty())
        return jointError(errs);

    // TODO very questionable
    m_shape3D = std::make_unique<TruncatedEllipsoidNet>(m_mean, m_mean, m_mean, 2.0 * m_mean, 0.0);
    return "";
}
