//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/HardParticle/Pyramid2.cpp
//! @brief     Implements class Pyramid2.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/HardParticle/Pyramid2.h"
#include "Base/Math/Constants.h"
#include "Base/Math/Functions.h"

const ff::PolyhedralTopology Pyramid2::topology = {{{{3, 2, 1, 0}, true},
                                                    {{0, 1, 5, 4}, false},
                                                    {{1, 2, 6, 5}, false},
                                                    {{2, 3, 7, 6}, false},
                                                    {{3, 0, 4, 7}, false},
                                                    {{4, 5, 6, 7}, true}},
                                                   false};

Pyramid2::Pyramid2(const std::vector<double> P)
    : IFormFactorPolyhedron(P)
    , m_length(m_P[0])
    , m_width(m_P[1])
    , m_height(m_P[2])
    , m_alpha(m_P[3])
{
    validateOrThrow();
}

Pyramid2::Pyramid2(double length, double width, double height, double alpha)
    : Pyramid2(std::vector<double>{length, width, height, alpha})
{
}

std::string Pyramid2::validate() const
{
    std::vector<std::string> errs;
    requestGt0(errs, m_length, "length");
    requestGt0(errs, m_width, "width");
    requestGt0(errs, m_height, "height");
    const double cot_alpha = Math::cot(m_alpha);
    if (m_alpha<= 0 || !std::isfinite(cot_alpha) || cot_alpha < 0)
        errs.push_back("pyramid angle alpha " + std::to_string(m_alpha) + " out of bounds");
    if (!errs.empty())
        return jointError(errs);

    double r = cot_alpha * 2 * m_height / m_length;
    double s = cot_alpha * 2 * m_height / m_width;
    if (r > 1)
        errs.push_back("parameters violate condition 2 * ctg(alpha) <= m_height / m_length");
    if (s > 1)
        errs.push_back("parameters violate condition 2 * ctg(alpha) <= m_height / m_width");
    if (!errs.empty())
        return jointError(errs);

    double D = m_length / 2;
    double d = m_length / 2 * (1 - r);
    double W = m_width / 2;
    double w = m_width / 2 * (1 - s);

    double zcom =
        m_height * (.5 - (r + s) / 3 + r * s / 4) / (1 - (r + s) / 2 + r * s / 3); // center of mass

    setPolyhedron(topology, -zcom,
                  {// base:
                   {-D, -W, -zcom},
                   {D, -W, -zcom},
                   {D, W, -zcom},
                   {-D, W, -zcom},
                   // top:
                   {-d, -w, m_height - zcom},
                   {d, -w, m_height - zcom},
                   {d, w, m_height - zcom},
                   {-d, w, m_height - zcom}});

    m_validated = true;
    return "";
}
