//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sample/HardParticle/TruncatedSpheroid.cpp
//! @brief     Implements class TruncatedSpheroid.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "Sample/HardParticle/TruncatedSpheroid.h"
#include "Base/Math/Bessel.h"
#include "Base/Math/Constants.h"
#include "Base/Math/IntegratorGK.h"
#include "Base/Util/Assert.h"
#include "Sample/Shapes/TruncatedEllipsoidNet.h"
#include <limits>

TruncatedSpheroid::TruncatedSpheroid(const std::vector<double> P)
    : IFormFactor(P)
    , m_radius(m_P[0])
    , m_height(m_P[1])
    , m_fp(m_P[2])
    , m_dh(m_P[3])
{
    validateOrThrow();
}

TruncatedSpheroid::TruncatedSpheroid(double radius, double height, double height_flattening,
                                     double dh)
    : TruncatedSpheroid(std::vector<double>{radius, height, height_flattening, dh})
{
}

complex_t TruncatedSpheroid::formfactor_at_bottom(C3 q) const
{
    ASSERT(m_validated);
    double H = m_height;
    double R = m_radius;
    double fp = m_fp;

    if (std::abs(q.mag()) <= std::numeric_limits<double>::epsilon())
        return M_PI / 3. / fp * (H * H * (3. * R - H / fp) - m_dh * m_dh * (3. * R - m_dh / fp));

    complex_t z_part = exp_I(q.z() * (H - fp * R));
    return M_TWOPI * z_part
           * ComplexIntegrator().integrate(
               [=](double Z) {
                   double R = m_radius;
                   double fp = m_fp;

                   double Rz = std::sqrt(R * R - Z * Z / (fp * fp));
                   complex_t qrRz = std::sqrt(q.x() * q.x() + q.y() * q.y()) * Rz;
                   complex_t J1_qrRz_div_qrRz = Math::Bessel::J1c(qrRz);

                   return Rz * Rz * J1_qrRz_div_qrRz * exp_I(q.z() * Z);
               },
               fp * R - H, fp * R - m_dh);
}

std::string TruncatedSpheroid::validate() const
{
    std::vector<std::string> errs;
    requestGt0(errs, m_radius, "radius");
    requestGt0(errs, m_height, "height");
    requestGt0(errs, m_fp, "height_flattening");
    requestGe0(errs, m_dh, "dh");
    if (m_height > 2 * m_radius * m_fp)
        errs.push_back("parameters violate condition H>2R*flattening");
    if (m_dh > m_height)
        errs.push_back("parameters violate condition dH<=H");
    if (!errs.empty())
        return jointError(errs);

    m_shape3D = std::make_unique<TruncatedEllipsoidNet>(
        m_radius, m_radius, m_fp * m_radius, m_height, m_dh);

    m_validated = true;
    return "";
}
